/**********************************************************************+
*
* SuperFP: A program for the reconstruction of the driving potential 
*          of stochastic trajectories, by solving an inverse problem
*          governed by the corresponding Fokker-Planck equation
*
**********************************************************************+
*
* THIS PACKAGE OF PROGRAM FILES IS A SUPPLEMENTAL MATERIAL FOR THE MANUSCRIPT 
* "Fokker-Planck analysis of superresolution microscopy images"
* BY MARIO ANNUNZIATO AND ALFIO BORZI' (NOV. 2023).
*
* THIS SOFTWARE IS EXPERIMENTAL AND IS RELEASED WITH ABSOLUTELY NO WARRANTY.
* IT PROVIDES THE NEED FOR CREATING THE EXECUTABLE FILES THAT HAVE BEEN USED
* BY THE AUTHORS FOR THE NUMERICAL EXPERIMENTS RECONSTRUCTIONS OF SECT. 4 
* AND 5 OF THE PAPER. 
*
* IT INHERITS THE SAME RIGHTS OF THE PUBLISHED PAPER. 
* IF YOU USE THIS CODE OF ANY PART OF IT, YOU ARE PLEASED TO QUOTE THE ARTICLE
* AND THE CODE NAME SuperFP (NOV. 2023)
*
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/




#include "BDF_Chang_Cooper.hpp"



BDF_CC::BDF_CC(){};




BDF_CC::BDF_CC(Model* model, Grid* grid_domain):Chang_Cooper(model,grid_domain) {






 H = &(CC_matrix);






 domain->alfax = 2*domain->dt/domain->dx;
 domain->alfay = 2*domain->dt/domain->dy;





 ff = new mat(domain->NN,domain->Nt,fill::zeros);


 pp = new mat(domain->NN,domain->Nt,fill::zeros);

 forward_solution = cube(ff->begin(),domain->Nx,domain->Ny,domain->Nt,false);

 backward_solution = cube(pp->begin(),domain->Nx,domain->Ny,domain->Nt,false);


}



BDF_CC::~BDF_CC() {

 delete ff;
 delete pp;

}





void BDF_CC::setup_initial_condition(mat& f0) {


 f = vectorise(f0);
 f_1 = f;
 f1 = vec(size(f),fill::zeros);


};






void BDF_CC::create_initial_BDF_value() {

 int N_eul=20;

    ff->col(0) = f;

 for(int k=0 ; k<N_eul ; k++) {

    bool status = spsolve(f, speye(size(*H)) + *H/(2*N_eul), f, "superlu");
 if(status == false) { cout << "Error when solving the system of equation for the BDF\n" << endl; }


 }

 ff->col(1) = f;

};

void BDF_CC::create_terminal_BDF_value(cube* source) {

 int N_eul=20;




    pp->col(pp->n_cols-1) = f;






 mat sT(domain->NN,2);

    sT.col(0) = vectorise(source->slice(source->n_slices-1));
    sT.col(1) = vectorise(source->slice(source->n_slices-2));



    mat h;

    vec idx_NN = regspace(0,domain->NN-1);




    vec t2 ={0,1};
    vec i_tempi= linspace(0,1.-1./N_eul,N_eul);




    interp2(t2,idx_NN, sT, i_tempi, idx_NN, h);



 h*=-domain->dt/N_eul;

 for(int k=0 ; k<N_eul ; k++) {

    bool status = spsolve(f,speye(size(*H)) + (*H).t()/(2*N_eul) , f+h.col(k), "superlu");
 if(status == false) { cout << "Error when solving the system of equation for the BDF\n" << endl; }

 }

 pp->col(pp->n_cols-2) = f;





};



vec BDF_CC::calculate_next_time_step() {

 vec fnew;


    bool status = spsolve(fnew, 3*speye(size(*H)) + *H, 4*f - f_1 , "superlu");
 if(status == false) { cout << "Error when solving the system of equation for the BDF\n" << endl; }
 return fnew;

};


vec BDF_CC::calculate_previous_time_step(vec* h) {




 vec fnew;

    bool status = spsolve(fnew, 3*speye(size(*H)) + (*H).t() , 4*f - f_1 + *h, "superlu");
 if(status == false) { cout << "Error when solving the system of equation for the BDF\n" << endl; }
 return fnew;

};

void BDF_CC::compute_forward(mat& f0) {






 build_Chang_Cooper_matrix();



 setup_initial_condition(f0);


 if (check_normalization()) { cout << "PDF iniziale non normalizzata" << endl; };


 create_initial_BDF_value();



 for(uword k=2 ; k<domain->Nt ; k++) {


 f1 = calculate_next_time_step();


    f_1 = f ;
    f = f1;


    ff->col(k) = f;

    check_normalization(k);

 }

};




void BDF_CC::compute_backward(mat& fT, cube& source) {

 setup_initial_condition(fT);


 create_terminal_BDF_value(&source);

 vec h;

 for(uword k=3 ; k<=domain->Nt ; k++) {




  h = -vectorise(source.slice(domain->Nt-k+1))*2*domain->dt;

  f1 = calculate_previous_time_step(&h);


  f_1 = f ;
  f = f1;


  pp->col(domain->Nt-k) = f;

 }




};





int BDF_CC::check_normalization(int n) {


 int flag=0;
 double norma;



 norma = sum(f)*domain->dS;

 if (abs(1-norma)> 1e-6)
 {
  cout << "Warning: norm = " << norma << " at time " << domain->t(n) << endl;
  flag=1;
 }

 return flag;

};
