/**********************************************************************+
*
* SuperFP: A program for the reconstruction of the driving potential 
*          of stochastic trajectories, by solving an inverse problem
*          governed by the corresponding Fokker-Planck equation
*
**********************************************************************+
*
* THIS PACKAGE OF PROGRAM FILES IS A SUPPLEMENTAL MATERIAL FOR THE MANUSCRIPT 
* "Fokker-Planck analysis of superresolution microscopy images"
* BY MARIO ANNUNZIATO AND ALFIO BORZI' (NOV. 2023).
*
* THIS SOFTWARE IS EXPERIMENTAL AND IS RELEASED WITH ABSOLUTELY NO WARRANTY.
* IT PROVIDES THE NEED FOR CREATING THE EXECUTABLE FILES THAT HAVE BEEN USED
* BY THE AUTHORS FOR THE NUMERICAL EXPERIMENTS RECONSTRUCTIONS OF SECT. 4 
* AND 5 OF THE PAPER. 
*
* IT INHERITS THE SAME RIGHTS OF THE PUBLISHED PAPER. 
* IF YOU USE THIS CODE OF ANY PART OF IT, YOU ARE PLEASED TO QUOTE THE ARTICLE
* AND THE CODE NAME SuperFP (NOV. 2023)
*
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/



/* Class definition of the BDF numerical integrator associated
 * to the Chang-Cooper method
 */

#ifndef BDF_CC_H
#define BDF_CC_H
 
#include <armadillo>
#include "Chang_Cooper.hpp"   



using namespace arma;


// apply the BDF method as derived class of Chang-Cooper

class BDF_CC : public Chang_Cooper {
	
	
	public:
		
	cube forward_solution;  // numerical solution of the forward equaton. 
							// Matrix of size Nx, Ny, Nt, associated to ff
	

	
	cube backward_solution; // numerical solution of the adjpoint equaton. 
							// Matrix of size Nx, Ny, Nt, associated to pp
	
	
	// compute the forward solution for all times
	void compute_forward(mat& f0);


	// compute the adjoint solution for all times
	// fT: terminal condition fT(x,y), source: source term source(x,y,t) at the r.h.s
	void compute_backward(mat& fT,cube& source);
	

    // constructor
	BDF_CC(Model* model, Grid* grid_domain);
	BDF_CC();   
	
	// destructor
	~BDF_CC();
	
	
	private:
	
	mat *ff;    // pointer to the soluton, as an alias of 'forward_solution'
	

	mat *pp;    // pointer to the adjoint solution, as an alias of 'backward_solution'
	
	
	vec f;      // space domain expanded as a vector. Dimension NN = Nx*Ny
	
		
	vec f_1;    // as above for two time steps (BDF)			
	
	vec f1;		// as above for the next time step
	
	sp_mat *H;     // pointer to the Chang-Cooper matrix
	
		
	void create_initial_BDF_value();   // create the 2nd initial data for the BDF
	void create_terminal_BDF_value(cube* source); //create the 2nd final data fot the BDF
		
	vec calculate_next_time_step();   // calculate the solution for the next time step	
	
	vec calculate_previous_time_step(vec* h);   // calculate the solution for the previous time step
	
	void setup_initial_condition(mat& f0); // set the initial condotion for the BDF
	
	int check_normalization(int n=0);   // check for the normalization of the PDF at the time step n
	
};


#endif
