/**********************************************************************+
*
* SuperFP: A program for the reconstruction of the driving potential 
*          of stochastic trajectories, by solving an inverse problem
*          governed by the corresponding Fokker-Planck equation
*
**********************************************************************+
*
* THIS PACKAGE OF PROGRAM FILES IS A SUPPLEMENTAL MATERIAL FOR THE MANUSCRIPT 
* "Fokker-Planck analysis of superresolution microscopy images"
* BY MARIO ANNUNZIATO AND ALFIO BORZI' (NOV. 2023).
*
* THIS SOFTWARE IS EXPERIMENTAL AND IS RELEASED WITH ABSOLUTELY NO WARRANTY.
* IT PROVIDES THE NEED FOR CREATING THE EXECUTABLE FILES THAT HAVE BEEN USED
* BY THE AUTHORS FOR THE NUMERICAL EXPERIMENTS RECONSTRUCTIONS OF SECT. 4 
* AND 5 OF THE PAPER. 
*
* IT INHERITS THE SAME RIGHTS OF THE PUBLISHED PAPER. 
* IF YOU USE THIS CODE OF ANY PART OF IT, YOU ARE PLEASED TO QUOTE THE ARTICLE
* AND THE CODE NAME SuperFP (NOV. 2023)
*
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/


/* Calculate the pictures from 3 to 5 of the paper
 * the pictures are the calculated membrane potential */


#include <iostream>
#include <fstream>
#include <string>
#include <ctime>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#define ARMA_USE_HDF5 

#include <armadillo>


#include "model.hpp"
#include "Non_lin_conjug_grad.hpp"


using namespace std;
using namespace arma;



int main() 
{
	
	time_t date;
	
	time (&date);
	//cout << ctime(&date) << endl;
	
// string for files and folders name
//
string base_data_dir, data, data_file, base_dir, name_dir, tot_dir;

string save_name, save_name_img, save_name_dat, nlcg_param_fname;

string str_date;

//  FOLDER PATH WHERE INPUT DATA ARE LOCATED
base_data_dir = "./input_data/";   

// FILE NAME OF THE INPUT DATA. 
// HISTOGRAMS OF THE PARTICLE TRAJECTORIES IN THE HDF5 FORMAT
data = "histograms1000_T90.h5";


// complete file path name
data_file = base_data_dir + data;


// PARENT FOLDER NAME WHERE RESULTS WILL BE SAVED
base_dir = "./membrane_computation_results/";

str_date =  ctime(&date);
str_date.pop_back();   

std::size_t j = str_date.find_first_of(':');
while(j!=std::string::npos) {
	str_date[j]='-';
	j=str_date.find_first_of(':',j+1);	
}



// folder name where results will be saved
name_dir = data + '_' + str_date + '/';  // folder name with the actual time

// path of the folder name where results will be saved
tot_dir = base_dir + name_dir;

// BASE NAME OF THE FILE WITH THE RESULTS
save_name = "Np1000_T90";

//cout << tot_dir << endl;


// makes the folder for results
if (mkdir(tot_dir.c_str(),DEFFILEMODE)) {
	
	cerr << errno << strerror(errno) << endl;
	cout << "\n Can not create the output directory \n" << endl;
		
	cout << "\n Please create the folder:"  << base_dir << "\n" << endl;
	
	return errno;
	
}

// FILE NAME FOR THE PARAMETERS OF THE NON-LINEAR CONJUGATE GRADIENT ALGORITHM
nlcg_param_fname = "nlcg_params.dat";



// load input Monte Carlo data
cube MC_data;
if (!MC_data.load(data_file,auto_detect)) {
		
	printf("Failed to load data file %s\n",data_file.c_str());
	return 1;	
}

// this is in case of data format in a text file rather than HDF5
// modify according to your own data format
#ifndef ARMA_USE_HDF5 
uword nr = MC_data.n_rows;
MC_data.reshape(nr,nr,MC_data.n_cols/nr);
#endif



  /**** PARAMETERS OF THE SIMULATION AND THE  FOKKER-PLANCK MODEL *****/
  
  // MODEL DEFINITION
  double sig  = 0.70; // INTENSITY OF THE GAUSSIAN NOISE OF THE FOKKER-PLANCK EQ.
  double alfa = 1e-4; // WEIGHT OF THE NORM OF THE POTENTIAL IN THE FUNCT. OBJECTIVE
  double beta = 1e-4; // WEIGHT OF THE NORM OF THE GRADIENT OF THE POTENTIAL
  // in the paper is alfa=beta
  
  double   xi = 1.; // WEIGHT OF THE TERMINAL CONDITION IN THE FUNCT. OBJECTIVE
  
  /*** Instances of model class ***/
  
  // test di scrittura nella cartella di salvataggio
  std::ofstream save_par(tot_dir+save_name+".par");
  save_par << sig << endl << alfa << endl << beta << endl << xi << endl;
  save_par.close();
  


  Param param(sig,alfa,beta,xi);  // object for parametrs
  
  Model fokker_planck(param);   // object for fokker-planck
 
 /**** END OF DEFINITIONS FOR MODEL AND PARAMETERS ***/


  
  /***** NUMERICAL DOMAIN *****/
  
  // BOUNDARIES OF THE 2D NUMERICAL DOMAIN 
  double ax=-3; double bx=-ax; double ay=ax; double by=-ax;
      
  // TIME INTERVAL FOR THE RECONSTRUCTION
  double TT=90; // THIS VALUE IS THE SAME OF THE MONTE CARLO SIMULATION
        
  // GRID POINTS OF THE NUMERICAL DOMAIN
  uword Nx=101; uword Ny=Nx;
  
  uword Nt = MC_data.n_slices; // number of histograms time points 
   
  cout << Nt << endl;
  
  vec x_data = linspace(ax,bx,MC_data.n_rows);
   
  
  /***** INITIAL CONDITION *****/
    
  // Temporary object of numerical grid
  // 
  Grid *dominio_tot = new Grid(ax,bx,Nx,ay,by,Ny); 
    

  // f0 is set for the first MPC window only
  
  double mux=0; double muy=0;   // COORDINATES OF THE MEAN VALUE OF THE GAUSS DISTRIBUTION
  double s0=-1;  // STANDARD DEVIATION OF THE GAUSS DISTRIBUTION. IF NEGATIVE
				// THE INITIAL DISTRIBUTION WILL BE UNIFORM
    
  mat f0(dominio_tot->Nx,dominio_tot->Ny);   // dato iniziale

    
  if (s0<0) {
	  	  
	 f0.ones();  // if the std deviation is negative then f0 is uniform
  } else 
  {
	f0 = exp(-(  square(dominio_tot->X-mux) + square(dominio_tot->Y-muy) ) /(2*s0*s0));
	}
	
	// normalization
  f0 = f0/(accu(f0)*dominio_tot->dx*dominio_tot->dy);
  
  delete dominio_tot; 
   
 /****** END OF THE SETTING OF THE INITIAL CONDITION ******/
  
   
  
  /*** START THE RECONSTRUCTION ON THE WINDOWS OF THE MODEL 
   * PREDICTIVE CONTROL ***/
   
  uword Nt_seq = 5;    // NUMBER OF WINDOWS OF THE MPC
  
   
  // vector of simulation times
  vec t = linspace(0,TT,Nt);
  
  // index vector for the time windows of the Model Predictive Control
  vec Nt_idx = round(linspace(0,Nt-1,Nt_seq+1));    
  
  
  /********  ITERATES ON THE MPC WINDOWS ********/
  //
  for ( uword idx_seq=0; idx_seq < Nt_seq; idx_seq++) {
	  
	// idx_seq is the counter of MPC windows
	printf("MPC time window num. %llu of %llu \n",idx_seq+1,Nt_seq);
	  
	// window time horizon
	double T  = t(Nt_idx(idx_seq+1)) - t(Nt_idx(idx_seq));  
	 
	// number of time grid points of the numerical grid
    double NT = Nt_idx(idx_seq+1) - Nt_idx(idx_seq)+1; 

    
    // istance of the numerical grid for a single MPC window
    Grid dominio(ax,bx,Nx,ay,by,Ny,T,NT);
    
	  
  /*** ---- BUILD THE DESIRED PDF fd ---- ***/
  

  cube fd(dominio.Nx,dominio.Ny,dominio.Nt,fill::zeros); 	   
	   
	  
	
  // INIZIALIZZARE LA fd A PARTIRE DAL DATO SPERIMENTALE y
  // SI DEVONO INTERPOLARE I VALORI DI y
 

	   uword i_mpc=0 ;
	   for ( uword i=Nt_idx(idx_seq) ; i <= Nt_idx(idx_seq+1); i++ ) {
		   
	   // i is the grid index inside the MPC window
	   
		    
	   // i_mpc is the index for the times of the MPC window
	     

	   // the numerical grid is usually finer of that the histograms
	   // hence it needs to be interpolated
	   //
	   interp2(x_data,x_data,MC_data.slice(i),dominio.y,dominio.x,fd.slice(i_mpc));
	 
	   //  fd  must be normalized
	   fd.slice(i_mpc) = fd.slice(i_mpc)/(accu(fd.slice(i_mpc))*dominio.dS); 
	   
	   
	   i_mpc++;
	   
   } 
   
   /*** END OF CONSTRUCTION OF fd ***/

	  // print the size of fd for a control
      cout << size(fd) << endl;
   

 
	   
  
  /**** DEFINE THE CONTROLLER VARIABLE u ****/
      
  mat u(size(f0));    	// u = control
  u.zeros(); 			// the initial value of the control is zero
				 
  
  /** END OF CONTROL DECLARATION ***/
  
 
  
  // CHECK FOR THE NORMALIZATION AND THE NON-NEGATIVITY OF THE INITIAL PDF
  if (f0.min()<0) {
   printf("The initial PDF f0 has the minimum negative value %g\n",f0.min());
  }
  
  printf("The normalization error of the initial PDF f0 is %g\n",accu(f0)*dominio.dS-1);
  

  
  /**** ALL THE OPTIMIZATION DATA PROBLEM ARE INITIALIZED ***/
 
 
  /******************************************/
  /*** PREPARE FOR THE START OF THE NLCG  ***/
  
  /**** INSTANCE FOR THE RESOLUTION OBJECT ****/

  // object for resolve the optimal control problem
  // u will be filled with the solution
  //
  NLCG nlcg(&fokker_planck,&dominio,f0,&u,&fd); 
  
  
  // load non-default parameters value for the NLCG method
  nlcg.load_params(nlcg_param_fname.c_str());
  
  // print parameters on the screen
  nlcg.print_params();
  
  uword lam0=10;    // INITIAL STEP FOR A SINGLE CICLE OF THE NLCG
  uword kmax=12;    // MAX NUMBER OF ITERATIONS FOR A SINGLE NLCG CICLE
  uword max_restart=10;  // MAX NUMBER OF RESTART SEQUENCES
    
    
  // FINALLY START THE NLCG METHOD   
  nlcg.start_restart_sequence(lam0,kmax,max_restart);
  
  /** HERE u IS FILLED WITH THE CALCULATED POTENTIAL **/
  
  
  // set f0 for the next MPC window with the last value of the 
  // PDF of the last MPC window
  f0 = nlcg.forward_solution.slice(dominio.Nt-1);
  
  
  
  /*** SOME POST PROCESSING FOR THE CALCULATED POTENTIAL ***/

  
  double min_u = min(vectorise(u));
  double max_u = max(vectorise(u));
  
  // the gray scale values are in between 0 and 255
  // normalize the max to 1 and multiply by 255
  u = 255*(u-min_u)/(max_u-min_u);
  
  u = flipud(u);   // flip the picture vertically
  
  
  // save picture and data of the actual MPC window
  // 
  save_name_img = save_name + '_' + std::to_string(idx_seq+1) + ".pgm"; 
  save_name_dat = save_name + '_' + std::to_string(idx_seq+1) + ".dat"; 
  
   
  if (!u.save(tot_dir+save_name_img, pgm_binary))
    cout << "Error: can not save " << save_name_img << endl;

  
  // mind that the potential u is saved with the scale factor 255
  if (!u.save(tot_dir+save_name_dat, raw_ascii))
	cout << "Error: can not save " << save_name_dat << endl;

  
  }  
    
   
    /***** FINE DEL CICLO SULLE FINESTRE MPC *****/
  
  // exit regularly from the program
  return 0;
  
};
