/**********************************************************************+
*
* SuperFP: A program for the reconstruction of the driving potential 
*          of stochastic trajectories, by solving an inverse problem
*          governed by the corresponding Fokker-Planck equation
*
**********************************************************************+
*
* THIS PACKAGE OF PROGRAM FILES IS A SUPPLEMENTAL MATERIAL FOR THE MANUSCRIPT 
* "Fokker-Planck analysis of superresolution microscopy images"
* BY MARIO ANNUNZIATO AND ALFIO BORZI' (NOV. 2023).
*
* THIS SOFTWARE IS EXPERIMENTAL AND IS RELEASED WITH ABSOLUTELY NO WARRANTY.
* IT PROVIDES THE NEED FOR CREATING THE EXECUTABLE FILES THAT HAVE BEEN USED
* BY THE AUTHORS FOR THE NUMERICAL EXPERIMENTS RECONSTRUCTIONS OF SECT. 4 
* AND 5 OF THE PAPER. 
*
* IT INHERITS THE SAME RIGHTS OF THE PUBLISHED PAPER. 
* IF YOU USE THIS CODE OF ANY PART OF IT, YOU ARE PLEASED TO QUOTE THE ARTICLE
* AND THE CODE NAME SuperFP (NOV. 2023)
*
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/

#include <armadillo>
#include "Optim_problem.hpp"
#include "Non_lin_conjug_grad.hpp"


using namespace arma;







NLCG::NLCG(Model* md, Grid* grid, mat &f0, mat *u, cube *p_fd):Optim_problem(md,grid,p_fd)
{

   u0=u;
   this->f0 = f0;

}




int NLCG::set_param(char *name_val, double val) {

 int found=0;

 for( int m=0 ; m < NUM_PARAMS ; m++ ) {

  if (!strcmp(name_val,par_name[m])) {

   switch (m) {

    case 0:
     bc = val;
     break;

    case 1:
     l_max = (uword) val;
     break;

    case 2:
     gamma = val;
     break;

    case 3:
     epsconv = val;
     break;

    case 4:
     tol_u = val;
     break;

    case 5:
     eps_pdf = val;
     break;

    }

   found = 1;
   break;

   }

  }

 return found;
}






int NLCG::load_params(const char *namef) {

 int n_load=0;
 int err=0;
 FILE *stream;

 char temp_name[32];
 double temp_val;



 if ((stream = fopen(namef, "r")) == NULL)
 {
  printf("Can not open the file %s\n", namef);

 } else {

  int line=1;

  printf("\n Reading file %s .... \n",namef);


  do {

   if (err) printf("read error at line %i\n",line);

   if(fscanf(stream, " %s %lf",temp_name,&temp_val)==2){

    if(set_param(temp_name,temp_val)) {
      n_load++;
     } else {

     printf("\n%s = %g not acknowledged\n",temp_name,temp_val);
     }

    err=0;

   }
   else { err=1; }

   line++;

  } while(!feof(stream));



 }


 if(fclose(stream)) { printf("\n Error in closing the file\n"); }
 else {printf(" done.\n");}

 return n_load;

}


void NLCG::print_params() {

 printf("\n%s %g", B_TRK , bc );
 printf("\n%s %lli", MAX_LEV , l_max );
 printf("\n%s %g", G_ARMIJO , gamma );
 printf("\n%s %g", E_CONV , epsconv );
 printf("\n%s %g", U_TOL , tol_u );
 printf("\n%s %g", E_PDF , eps_pdf );
 printf("\n");

}




double NLCG::inner(mat& v, mat& w)
{

 return dot(v,w)*domain->dS;

}



double NLCG::my_norm(mat& v)
{

 return norm(vectorise(v))*sqrt(domain->dS);

}


double NLCG::my_norm_diff(mat &v1, mat &v2)
{
 mat v = v1-v2;

 return my_norm(v);
}







uword NLCG::start_restart_sequence(double lam0, uword kmax, uword max_restart)
{




 uword precision_error;

 uword restart = 1;





 J = calc_forward_and_objective(f0,*u0);


 cout << "J = " << J << endl;

 gr0 = calc_objective_gradient();

 grdnorm0 = my_norm(gr0);

 cout << "grdnorm0 = " << grdnorm0 << endl;




 while (restart <= max_restart) {

  precision_error = 0;

  printf(" restart sequence %lli\n", restart);





  if (start_nlgc_sequence(precision_error,lam0,kmax)==1) break;

  restart++;

    printf("\n Restart of the line search. Iteration %lli\n",restart);

  precision_error = 1;

 }

 return precision_error;
}





uword NLCG::start_nlgc_sequence(uword& precision_error, double& lam0, uword& kmax) {



 mat ds = -gr0;

 g_proj = inner(gr0,ds);


    double gnorm = my_norm(gr0);

    printf("gradient norm = %g\n",gnorm);

 double lam = lam0;

 uword i=1;


   while ( i <= kmax && gnorm > epsconv*grdnorm0 ) {


  if (i>1) {



   lam=3*my_norm_diff(*u0,u_old)/gnorm;
   lam=min(lam,lam0);

  }


  printf(" nlcg sequence %lli \n",i);


  if (nlcg_step_down_Dai_Yuan(lam,ds) > 1) break;

  gnorm = my_norm(gr0);

  i++;
 }

 return i;
}






uword NLCG::nlcg_step_down_Dai_Yuan(double &lam, mat& ds) {

 uword err;

 double beta;

 double J1;

 double norm_u = my_norm(*u0);

 printf("\n J=%20.16g \n",J);

 K_J++;
 JJ.resize(K_J);
 JJ(K_J-1)=J;

 cout << J << endl;



 err = Armijo_line_search(lam,ds,J1,norm_u);

  if (err > 1) {





  J = calc_forward_and_objective(f0,*u0);


 switch(err) {

  case 2:
  printf("\n The maximum divided up level of the Armijo's step length has been reached\n");
  break;

  case 3:
  printf("\n The step lenght is vanished \n");
  break;

  case 4:
  printf("\n The tolerance of the control has been reached, but the step lenght has not been found \n");

 }



  } else {






   mat gr1 = calc_objective_gradient();


   mat g_diff = gr1-gr0;

   beta = inner(gr1,gr1) / inner(g_diff,ds);


   ds = -gr1 + beta*ds;

   J = J1;

   gr0=gr1;
 }


   return err;
}





uword NLCG::Armijo_line_search(double &lam, mat &ds, double& J1, double norm_u) {

 uword err=0;

 mat &u = *u0;

 double dJ;

 double err_toll=1;

 uword l=1;



 double norm_ds = my_norm(ds);


    while ( norm_ds*lam > tol_u*norm_u && l<l_max) {

  u_new = u + lam*ds;




  J1 = calc_forward_and_objective(f0,u_new);

  dJ = J1 - J;



  printf("\n lambda step=%g,  level %lli, J1=%g, lam*ds=%g \n",lam,l,J1, norm_ds*lam);

  if (dJ <= lam*gamma*g_proj) {



   u_old = u;
   u = u_new;

   err_toll=0;

   break;

  }

  lam = lam*bc;
  l++;

 }





      if (l >= l_max) {err=2;}
      else if ( lam==0 ) {err=3;}
      else if (err_toll) {err=4;}

    return err;
}
