/**********************************************************************+
*
* SuperFP: A program for the reconstruction of the driving potential 
*          of stochastic trajectories, by solving an inverse problem
*          governed by the corresponding Fokker-Planck equation
*
**********************************************************************+
*
* THIS PACKAGE OF PROGRAM FILES IS A SUPPLEMENTAL MATERIAL FOR THE MANUSCRIPT 
* "Fokker-Planck analysis of superresolution microscopy images"
* BY MARIO ANNUNZIATO AND ALFIO BORZI' (NOV. 2023).
*
* THIS SOFTWARE IS EXPERIMENTAL AND IS RELEASED WITH ABSOLUTELY NO WARRANTY.
* IT PROVIDES THE NEED FOR CREATING THE EXECUTABLE FILES THAT HAVE BEEN USED
* BY THE AUTHORS FOR THE NUMERICAL EXPERIMENTS RECONSTRUCTIONS OF SECT. 4 
* AND 5 OF THE PAPER. 
*
* IT INHERITS THE SAME RIGHTS OF THE PUBLISHED PAPER. 
* IF YOU USE THIS CODE OF ANY PART OF IT, YOU ARE PLEASED TO QUOTE THE ARTICLE
* AND THE CODE NAME SuperFP (NOV. 2023)
*
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/


/* Definitions for the class for the non linear conjugate gradient
 */


#ifndef NON_LIN_CG
#define NON_LIN_CG

#include <armadillo>
#include "Optim_problem.hpp"

using namespace arma;

// definitions for parameters names end default values

#define B_TRK  			"BACK_TRACK"
#define B_TRK_DEF  		0.5
#define MAX_LEV			"MAX_LEVELS"
#define MAX_LEV_DEF 	10
#define G_ARMIJO		"GAMMA"
#define G_ARMIJO_DEF 	0.1
#define E_CONV			"EPS_CONV"
#define E_CONV_DEF		1e-4
#define U_TOL			"U_TOLERANCE"
#define U_TOL_DEF		1e-4
#define E_PDF			"EPS_PDF"
#define E_PDF_DEF		1e-8

#define NUM_PARAMS		6        // number of parameters that can be loaded
#define MAX_NAME_PAR_LENGTH 16   // maximum lenght of the parameter name


// The class inherits from the optimal problem class
// NLGC: acronym for Non Linear Conjugate Gradient

class NLCG: public Optim_problem {

    
    // name parameter list        
    const char par_name[NUM_PARAMS][MAX_NAME_PAR_LENGTH+1]=
    { B_TRK , MAX_LEV , G_ARMIJO , E_CONV , U_TOL , E_PDF };
           
               
	double bc = B_TRK_DEF;  	// backtracking constant

	uword l_max = MAX_LEV_DEF;  // max search level of steplength with the Armijo condition
	
	double gamma = G_ARMIJO_DEF;// constant for Armijo condition

	double epsconv = E_CONV_DEF;// epsilon constant for the iterative steplength search

	double tol_u = U_TOL_DEF;   // relative uncertainty of the control function u, that is
								// used in the stop criteria

								
	double eps_pdf = E_PDF_DEF; // threshold for the value of the PDF that force the control u to zero
								// (not used !) 
	
							
	double grdnorm0;			// initial value of the gradient norm (input)
		
	double g_proj;				// projection gr*ds;
				
	double J; 					// value of the objective functional	
	
	mat gr0;					// gradient		
				
	mat f0;						// PDF initial value (in local memory)
								
	mat *DFunc;			
	
	mat *u0;					// external pointer to the initial value of the control u 
					
	mat u_old;					// control value at the previous iteration step 

	mat u_new;					// control value at the next iteration step
	
	// set the value val to the parameter name_val
	int set_param(char *name_val, double val); 

	double inner(mat& v, mat& w);  	// inner product

	double my_norm(mat& v);			// L2 norm
	
	double my_norm_diff(mat &v1, mat &v2); // norm of v1-v2
	
	// descent step of the NLCG with the Dai-Yuan method
	uword nlcg_step_down_Dai_Yuan(double &lam, mat& ds);  
	
	// line search with the Armijo condition
	uword Armijo_line_search(double &lam, mat &ds, double& J1, double norm_u); 

	// PUBLIC METHODS

	public:
	
	
	vec JJ;				// keep the functional value J
	uword K_J = 0;		// overall counter ( restarts and NLCG iteration) 
						// used to account the J values for each iteration
					
	
	// constructors
	NLCG();
	
	// constructor to bind the model, numerical grid, initial condition, the (unknown) control
	// and the desired function
	//
	NLCG(Model* md, Grid* grid, mat &f0, mat *u, cube *p_fd);
	

	
	// *** THIS IS THE MAIN METHOD TO BE CALLED ***
	//    START THE SEARCH FOR MINIMIZING THE FUNCTIONAL WITH THE NLCG METHOD
	//
	// ARGUMENTS: initial step length, maximum number of iterations, maximum number of
	// restarts of the NLCG method
	//
	// RETURN: 0 in case the required precision has been accomplished
	//         1 otherwise
	//
	uword start_restart_sequence(double lam0, uword kmax, uword max_restart);
	
	

	// Start a single sequence of NLCG. It is included in the method start_restart_sequence().
	// It contains the method nlcg_step_down().
	// Return the number of iteration carried out for the descent sequence
	//
	uword start_nlgc_sequence(uword& precision_error, double& lam0, uword& kmax);
	
	int load_params(const char *namef); // load parameters from a file. return 0 when error
	
	void print_params();  // print the currents parameters
};


#endif
