/**********************************************************************+
*
* SuperFP: A program for the reconstruction of the driving potential 
*          of stochastic trajectories, by solving an inverse problem
*          governed by the corresponding Fokker-Planck equation
*
**********************************************************************+
*
* THIS PACKAGE OF PROGRAM FILES IS A SUPPLEMENTAL MATERIAL FOR THE MANUSCRIPT 
* "Fokker-Planck analysis of superresolution microscopy images"
* BY MARIO ANNUNZIATO AND ALFIO BORZI' (NOV. 2023).
*
* THIS SOFTWARE IS EXPERIMENTAL AND IS RELEASED WITH ABSOLUTELY NO WARRANTY.
* IT PROVIDES THE NEED FOR CREATING THE EXECUTABLE FILES THAT HAVE BEEN USED
* BY THE AUTHORS FOR THE NUMERICAL EXPERIMENTS RECONSTRUCTIONS OF SECT. 4 
* AND 5 OF THE PAPER. 
*
* IT INHERITS THE SAME RIGHTS OF THE PUBLISHED PAPER. 
* IF YOU USE THIS CODE OF ANY PART OF IT, YOU ARE PLEASED TO QUOTE THE ARTICLE
* AND THE CODE NAME SuperFP (NOV. 2023)
*
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/

#include "Optim_problem.hpp"
#include "gradients.hpp"



Optim_problem::Optim_problem() {

}


Optim_problem::Optim_problem(Model* md, Grid* grid,cube *p_fd):BDF_CC(md,grid) {




 fd = p_fd;





 Laplace_Neumann_mat = build_Laplace_Neumann_matrix();


}


Optim_problem::~Optim_problem() {

 if (Laplace_Neumann_mat != NULL) delete Laplace_Neumann_mat;

}





sp_mat* Optim_problem::build_Laplace_Neumann_matrix() {


 if (!domain->square_element_flag) return NULL;





 double h = domain->dx;
 uword NN = domain->NN;
 uword Nx = domain->Nx;
 uword Ny = domain->Ny;



 sp_mat* LN= new sp_mat(NN,NN);





 LN->diag( 0) += (4+h*h);
 LN->diag(+1) -= 1.;
 LN->diag(-1) -= 1.;

 LN->diag(+Nx)-= 1.;
 LN->diag(-Nx)-= 1.;

 vec w(NN,fill::zeros);
 w.head(Nx) -= 1.;

 LN->diag() += w;

 w.zeros();
 w.tail(Nx) -= 1.;
 LN->diag() += w;







 mat c(Nx,Ny,fill::zeros);
 c.row(c.n_rows-1) += 1.;







 LN->diag(-1) += resize(vectorise(c),c.n_elem-1,1);
 LN->diag(+1) += resize(vectorise(c),c.n_elem-1,1);


 LN->diag() -= vectorise(c);
 LN->diag() -= vectorise(flipud(c));


 return LN;

}




double Optim_problem::objective() {


 double J;


 mat *u = md->u;
 double dS = domain->dS;


 mat gUx = grad_x(u,domain->dx);
 mat gUy = grad_y(u,domain->dy);



 J = ( md->param.alfa*pow(norm(vectorise(*u)),2) +
          md->param.beta*(pow(norm(vectorise(gUx)),2) + pow(norm(vectorise(gUy)),2) ))*dS/2.;



 J = J + 0.5*pow(norm(vectorise(forward_solution-*fd)),2)*dS*domain->dt
    + 0.5*md->param.xi*pow(norm(vectorise(forward_solution.slice(forward_solution.n_slices-1) - (fd->slice(fd->n_slices-1)))),2)*dS;





 return J;

}

mat Optim_problem::calc_objective_gradient() {




 calc_backward();


 return grad_H1();

}

mat Optim_problem::grad_L2() {




 cube dpx = grad_x(&backward_solution,domain->dx);
 cube dpy = grad_y(&backward_solution,domain->dy);


 dpx = forward_solution % dpx;
 dpy = forward_solution % dpy;


 cube div1 = grad_x(&dpx,domain->dx);
 cube div2 = grad_y(&dpy,domain->dy);

 mat grad = - (mat) sum(div1+div2,2) * domain->dt + md->param.alfa * (*md->u)
      -md->param.beta * discrete_Laplace();


 return grad;
}


mat Optim_problem::discrete_Laplace() {




 return (domain->square_element_flag) ?
  ::discrete_Laplace(md->u,domain->dx) :
  ::discrete_Laplace(md->u,domain->dx,domain->dy);

}




mat Optim_problem::grad_H1() {


 mat g = grad_L2();


 vec gH1 = spsolve(*Laplace_Neumann_mat,vectorise(g)*domain->dx*domain->dx);


    return reshape(gH1,domain->Nx,domain->Ny);

}



double Optim_problem::calc_forward_and_objective(mat& f0, mat& u) {



 md->u = &u;

 compute_forward(f0);

 return objective();

}



void Optim_problem::calc_backward() {


 source = forward_solution - *fd;


 pT = - md->param.xi *
   (forward_solution.slice(forward_solution.n_slices-1) -
    fd->slice(fd->n_slices-1));


 compute_backward(pT,source);

}
