/**********************************************************************+
*
* SuperFP: A program for the reconstruction of the driving potential 
*          of stochastic trajectories, by solving an inverse problem
*          governed by the corresponding Fokker-Planck equation
*
**********************************************************************+
*
* THIS PACKAGE OF PROGRAM FILES IS A SUPPLEMENTAL MATERIAL FOR THE MANUSCRIPT 
* "Fokker-Planck analysis of superresolution microscopy images"
* BY MARIO ANNUNZIATO AND ALFIO BORZI' (NOV. 2023).
*
* THIS SOFTWARE IS EXPERIMENTAL AND IS RELEASED WITH ABSOLUTELY NO WARRANTY.
* IT PROVIDES THE NEED FOR CREATING THE EXECUTABLE FILES THAT HAVE BEEN USED
* BY THE AUTHORS FOR THE NUMERICAL EXPERIMENTS RECONSTRUCTIONS OF SECT. 4 
* AND 5 OF THE PAPER. 
*
* IT INHERITS THE SAME RIGHTS OF THE PUBLISHED PAPER. 
* IF YOU USE THIS CODE OF ANY PART OF IT, YOU ARE PLEASED TO QUOTE THE ARTICLE
* AND THE CODE NAME SuperFP (NOV. 2023)
*
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/

#include <armadillo>
#include "gradients.hpp"

using namespace arma;







mat grad_x(mat* f, double dx)
{


   mat g(size(*f));

   uword r_end = g.n_rows-1;


   g.rows(1,r_end-1) = ( f->rows(2,r_end) - f->rows(0,r_end-2) )/ (2*dx);


   g.row(0) = ( f->row(1) - f->row(0) ) / dx;
   g.row(r_end) = ( f->row(r_end) - f->row(r_end-1) ) / dx;

   return g;

}





mat grad_y(mat* f, double dx)
{


   mat g(size(*f));

   uword c_end = g.n_cols-1;


   g.cols(1,c_end-1) = ( f->cols(2,c_end) - f->cols(0,c_end-2) )/ (2*dx);


   g.col(0) = ( f->col(1) - f->col(0) ) / dx;
   g.col(c_end) = ( f->col(c_end) - f->col(c_end-1) ) / dx;

   return g;

}



mat discrete_Laplace(mat* f_ptr, double dx) {

  mat& f = *f_ptr;

 mat g(size(f),fill::zeros);

 uword r_end = g.n_rows-1;
 uword c_end = g.n_cols-1;



 if (r_end < 3 || c_end < 3 )
 {
  return g;
 }



  g(span(1,r_end-1),span(1,c_end-1)) =

   ( f(span(2,r_end ),span(1,c_end-1)) +
   f(span(0,r_end-2),span(1,c_end-1)) )/ dx +
   ( f(span(1,r_end-1),span(2,c_end )) +
   f(span(1,r_end-1),span(0,c_end-2)) )/ dx -
      f(span(1,r_end-1),span(1,c_end-1))*4/dx;







 g(span(0,0),span(1,c_end-1)) =

    ( -5*f(span(1,1),span(1,c_end-1))
   +4*f(span(2,2),span(1,c_end-1))
   -1*f(span(3,3),span(1,c_end-1))
   +
      f(span(0,0),span(2,c_end )) +
   f(span(0,0),span(0,c_end-2)) ) / dx;


 g(span(r_end,r_end),span(1,c_end-1)) =

    ( -5*f(span(r_end-1,r_end-1),span(1,c_end-1))
   +4*f(span(r_end-2,r_end-2),span(1,c_end-1))
   -1*f(span(r_end-3,r_end-3),span(1,c_end-1))
   +
      f(span(r_end,r_end),span(2,c_end )) +
   f(span(r_end,r_end),span(0,c_end-2)) )/ dx;


 g(span(1,r_end-1),span(0,0)) =

    ( -5*f(span(1,r_end-1),span(1,1))
   +4*f(span(1,r_end-1),span(2,2))
   -1*f(span(1,r_end-1),span(3,3))
   +
      f(span(2,r_end ),span(0,0)) +
   f(span(0,r_end-2),span(0,0)) )/ dx;



 g(span(1,r_end-1),span(c_end,c_end)) =

    ( -5*f(span(1,r_end-1),span(c_end-1,c_end-1))
   +4*f(span(1,r_end-1),span(c_end-2,c_end-2))
   -1*f(span(1,r_end-1),span(c_end-3,c_end-3))
   +
      f(span(2,r_end ),span(c_end,c_end)) +
   f(span(0,r_end-2),span(c_end,c_end)) )/ dx;



 g(0,0) = (2*f(0,0)-5*f(1,0)+4*f(2,0)-f(3,0) +
     2*f(0,0)-5*f(0,1)+4*f(0,2)-f(0,3))/dx;

 g(0,c_end) = (2*f(0,c_end)-5*f(1,c_end)+4*f(2,c_end)-f(3,c_end) +
    2*f(0,c_end)-5*f(0,c_end-1)+4*f(0,c_end-2)-f(0,c_end-3))/dx;

 g(r_end,0) = (2*f(r_end,0)-5*f(r_end-1,0)+4*f(r_end-2,0)-f(r_end-3,0) +
         2*f(r_end,0)-5*f(r_end,1)+4*f(r_end,2)-f(r_end,3))/dx;

 g(r_end,c_end) =
    (2*f(r_end,c_end)-5*f(r_end-1,c_end)+4*f(r_end-2,c_end)-f(r_end-3,c_end) +
        2*f(r_end,c_end)-5*f(r_end,c_end-1)+4*f(r_end,c_end-2)-f(r_end,c_end-3))/dx;


  return g;


}




mat discrete_Laplace(mat* f_ptr, double dx, double dy)
{
 mat& f = *f_ptr;

 mat g(size(f),fill::zeros);

 uword r_end = g.n_rows-1;
 uword c_end = g.n_cols-1;



 if (r_end < 3 || c_end < 3 )
 {
  return g;
 }



  g(span(1,r_end-1),span(1,c_end-1)) =

   ( f(span(2,r_end ),span(1,c_end-1)) +
   f(span(0,r_end-2),span(1,c_end-1)) )/ dx +
   ( f(span(1,r_end-1),span(2,c_end )) +
   f(span(1,r_end-1),span(0,c_end-2)) )/ dy -
      f(span(1,r_end-1),span(1,c_end-1))*2*(dx+dy)/(dx*dy);







 g(span(0,0),span(1,c_end-1)) =

    ( -5*f(span(1,1),span(1,c_end-1))
   +4*f(span(2,2),span(1,c_end-1))
   -1*f(span(3,3),span(1,c_end-1)) ) / dx
   +
    ( f(span(0,0),span(2,c_end )) +
   f(span(0,0),span(0,c_end-2)) ) / dy +
   f(span(0,0),span(1,c_end-1))*2*(-dx+dy)/(dx*dy);


 g(span(r_end,r_end),span(1,c_end-1)) =

    ( -5*f(span(r_end-1,r_end-1),span(1,c_end-1))
   +4*f(span(r_end-2,r_end-2),span(1,c_end-1))
   -1*f(span(r_end-3,r_end-3),span(1,c_end-1)) ) / dx
   +
    ( f(span(r_end,r_end),span(2,c_end )) +
   f(span(r_end,r_end),span(0,c_end-2)) )/ dy +
   f(span(r_end,r_end),span(1,c_end-1))*2*(-dx+dy)/(dx*dy);


 g(span(1,r_end-1),span(0,0)) =

    ( -5*f(span(1,r_end-1),span(1,1))
   +4*f(span(1,r_end-1),span(2,2))
   -1*f(span(1,r_end-1),span(3,3)) ) / dy
   +
    ( f(span(2,r_end ),span(0,0)) +
   f(span(0,r_end-2),span(0,0)) )/ dx +
   f(span(1,r_end-1),span(0,0))*2*(dx-dy)/(dx*dy);



 g(span(1,r_end-1),span(c_end,c_end)) =

    ( -5*f(span(1,r_end-1),span(c_end-1,c_end-1))
   +4*f(span(1,r_end-1),span(c_end-2,c_end-2))
   -1*f(span(1,r_end-1),span(c_end-3,c_end-3)) ) / dy
   +
    ( f(span(2,r_end ),span(c_end,c_end)) +
   f(span(0,r_end-2),span(c_end,c_end)) )/ dx +
   f(span(1,r_end-1),span(c_end,c_end))*2*(dx-dy)/(dx*dy);



 g(0,0) = (2*f(0,0)-5*f(1,0)+4*f(2,0)-f(3,0))/dx +
    (2*f(0,0)-5*f(0,1)+4*f(0,2)-f(0,3))/dy;

 g(0,c_end) = (2*f(0,c_end)-5*f(1,c_end)+4*f(2,c_end)-f(3,c_end))/dx +
    (2*f(0,c_end)-5*f(0,c_end-1)+4*f(0,c_end-2)-f(0,c_end-3))/dy;

 g(r_end,0) = (2*f(r_end,0)-5*f(r_end-1,0)+4*f(r_end-2,0)-f(r_end-3,0))/dx +
        (2*f(r_end,0)-5*f(r_end,1)+4*f(r_end,2)-f(r_end,3))/dy;

 g(r_end,c_end) =
    (2*f(r_end,c_end)-5*f(r_end-1,c_end)+4*f(r_end-2,c_end)-f(r_end-3,c_end))/dx +
       (2*f(r_end,c_end)-5*f(r_end,c_end-1)+4*f(r_end,c_end-2)-f(r_end,c_end-3))/dy;



  return g;
}

cube grad_x(cube* f, double dx)
{


   cube g(size(*f));

   uword r_end = g.n_rows-1;


   g.rows(1,r_end-1) = ( f->rows(2,r_end) - f->rows(0,r_end-2) )/ (2*dx);


   g.row(0) = ( f->row(1) - f->row(0) ) / dx;
   g.row(r_end) = ( f->row(r_end) - f->row(r_end-1) ) / dx;

   return g;

}


cube grad_y(cube* f, double dx)
{


   cube g(size(*f));

   uword c_end = g.n_cols-1;


   g.cols(1,c_end-1) = ( f->cols(2,c_end) - f->cols(0,c_end-2) )/ (2*dx);


   g.col(0) = ( f->col(1) - f->col(0) ) / dx;
   g.col(c_end) = ( f->col(c_end) - f->col(c_end-1) ) / dx;

   return g;

}
