/**********************************************************************+
*
* SuperFP: A program for the reconstruction of the driving potential 
*          of stochastic trajectories, by solving an inverse problem
*          governed by the corresponding Fokker-Planck equation
*
**********************************************************************+
*
* THIS PACKAGE OF PROGRAM FILES IS A SUPPLEMENTAL MATERIAL FOR THE MANUSCRIPT 
* "Fokker-Planck analysis of superresolution microscopy images"
* BY MARIO ANNUNZIATO AND ALFIO BORZI' (NOV. 2023).
*
* THIS SOFTWARE IS EXPERIMENTAL AND IS RELEASED WITH ABSOLUTELY NO WARRANTY.
* IT PROVIDES THE NEED FOR CREATING THE EXECUTABLE FILES THAT HAVE BEEN USED
* BY THE AUTHORS FOR THE NUMERICAL EXPERIMENTS RECONSTRUCTIONS OF SECT. 4 
* AND 5 OF THE PAPER. 
*
* IT INHERITS THE SAME RIGHTS OF THE PUBLISHED PAPER. 
* IF YOU USE THIS CODE OF ANY PART OF IT, YOU ARE PLEASED TO QUOTE THE ARTICLE
* AND THE CODE NAME SuperFP (NOV. 2023)
*
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*/


/** Support functions for the model class 
 */

#ifndef MODEL_H
#define MODEL_H

#include<armadillo>
#include<iostream>


using namespace std;
using namespace arma;




/* 
 * Grid :  structure with the definition of the square mesh grid
 *
 * 
 * mat x_y:    domain grid: x along the coloumns, y along the rows 
 *             i.e. x_y(x,y), where x is row index, y coloumn index 
 * 			   The number of elements is Nx * Ny	
 * 
 * mat xx_y:  mesh grid along x. Dimension is (Nx-1) * Ny
 * 
 * mat x_yy:  mesh grid along y. Dimension is Nx * (Ny-1) 
 *  
 * 
 * vec x:  x-grid vector
 * vec y:  y-grid vector
 * 
 * vec xx: grid vector for the boundary of the cell along x
 * vec yy: grid vector for the boundary of the cell along y
 *  
 * 
 * 
 * Nx: number of the mesh points for the x dimension.
 * Ny: number of the mesh points for the y dimension.
 * 
 * dx: mesh step size x
 * dy: mesh step sixe y
 * 
 * 
*/

//  Grid struct for the domain numerical grid 
struct Grid {
	
	double prec = 1e-15;
	
	///<  boundaries of the rectangular domain
	
	double ax; double bx;  
	double ay; double by;
		
	double T; 			   ///< final time
		
	uword Nx; uword Ny;  ///< number of grid points for each dimension
	
	uword NN;   	///< NN=Nx*Ny total number of the 2-dim grid
	
			 
    uword Nt;	 ///< size of the time grid points 
    
    uword NNN;   ///< total number of grid points of the space-time domain						
	
	 vec x ; ///< vector grid of x
	 vec y ; ///< vector grid of y
 
	 vec xx; ///< vector grid of the x cell boundaries
	 vec yy; ///< vector grid of the y cell boundaries
	 
	 mat X;  ///< matrix of the repetition of the x grid
	 mat Y;  ///< matrix of the repetition of the x grid 
	 
	 vec t ; ///< time grid
	 
	 double dx; double dy;  ///< size of the space grid
	 double dS;				///< Surface element
	 double dt;				///< time step
		
	 uword square_element_flag=0; // is set to 1 if dx=dy inside the precision 'prec'
		
	 double alfax=0;  double alfay=0;  // dx/dt 
	
	// base constructor
	Grid() {};
	
	
	/* build the numerical grid. [A,B] interval, N number of grid points
	 * 1 = x-axis, 2 = y-axis
	 */
	Grid(double A1, double B1, uword N1, double A2, double B2, uword N2, double Tf=0, uword NT=2) 
	{
		
		ax=A1; bx=B1; Nx=N1;
		ay=A2; by=B2; Ny=N2;
		
		NN = Nx*Ny;     // total number of grid points of the domain
						// the number of elements of the CC matrix will 
						// be NN*NN
		
			
		// vectors of the grid points 
		x = linspace(ax,bx,Nx);		
		y = linspace(ay,by,Ny);	
		
		dx = x(1)-x(0);
		dy = y(1)-y(0);
		
		// 
		if (abs(dx-dy)< prec) square_element_flag = 1;
		else {
			printf("\n This version does not support dx not equals to dy. Exit \n");
			exit(1);
		};
		
		dS = dx*dy;    // area differential
		
		// vector with the mesh point related to the cell boundaries (inner to the domain)
		xx = linspace(ax+dx/2.,bx-dx/2.,Nx-1);
		yy = linspace(ay+dy/2.,by-dy/2.,Ny-1);			
		
		
		// matrices for the evaluation of the 2D functions
		X = repmat(x,1,Ny);
		Y = repmat(y.t(),Nx,1);
		
		
		// time vector
		T=Tf;
		Nt=NT;
		
		t = linspace(0,T,Nt);
		dt = t(1)-t(0);
		
		// total grid points for the space-time domain
		NNN = NN*Nt;
		
	};
	
	
	
};

// ----------- END OF DEFINITION OF THE GRID CLASS ----------------



// DEFINITION OF THE PARAMETER CLASS

/**
 * Class of parameters
 */

struct Param {
	
	double sig;  ///< intensity of the gaussian noise
	
	double alfa; ///< weight of the norm in the cost function
		
	double beta; ///< weight of the norm of gradient in the cost function	
	
	double xi;	///< weight of the norm the terminal condition in the cost function
	

	// constructors
	Param() {};
	
	Param(double s, double alf, double bet, double xi_) {
		
		sig = s;	
		alfa=alf;
		beta=bet;
		xi = xi_;

		
	};
	
	
};

/* --- END OF DEFINITION OF THE PARAMETER CLASS --- */


// Definition of the model class

/** Class of the model equation
 */
class Model {
	
	public:
	
	// model parameters
	Param param;
	
	// pointer to the matrix u of the control function u(x,y)
	mat* u;	
	

	// constructor (parameters and control)
	//
	Model(Param& p, mat* u_ptr=NULL) { set_Param(p); u = u_ptr; };

	///< Set the parameters values
	void set_Param(Param& p) { param = p; };
	

	///< Get the parameters values
	Param get_Param(void) { return param; };

	
	///< Sigma function of the noise term in the Ito equation: sig(x)*dW
	/// nx and ny are matrix sizes of the domain, sig is the coefficient
	mat sigma(uword nx, uword ny, double sig );
	
	

	// calculates the values on the model equation
	// for the CC method (see references)
	
	// ritorna i valori di B lungo x
	mat val_Bx(Grid* domain, mat* uu);	
	
	// ritorna i valori di B lungo y
	mat val_By(Grid* domain, mat* uu);
	
	// ritorna i valori di C con bordo cella lungo x
	mat val_Cx(Grid* domain);
	
	// ritorna i valori di C con bordo cella lungo y
	mat val_Cy(Grid* domain);
		
	
};


#endif
